;;;; hato-base64.scm -- permissive RFC3548 base64 implementation
;;
;; Copyright (c) 2005-2008 Alex Shinn.  All rights reserved.
;; BSD-style license: http://synthcode.com/license.txt

;; Procedure: base64-encode-string str
;;   Return a base64 encoded representation of string according to the
;;   official base64 standard as described in RFC3548.

;; Procedure: base64-decode-string str
;;   Return a base64 decoded representation of string, also interpreting
;;   the alternate 62 & 63 valued characters as described in RFC3548.
;;   Other out-of-band characters are silently stripped, and = signals
;;   the end of the encoded string.  No errors will be raised.

;; Procedure: base64-encode [port]
;; Procedure: base64-decode [port]
;;   Variations of the above which read and write to ports.

;; Procedure: base64-encode-header enc str [start-col max-col nl]
;;   Return a base64 encoded representation of string as above,
;;   wrapped in =?ENC?B?...?= as per RFC1522, split across multiple
;;   MIME-header lines as needed to keep each lines length less than
;;   MAX-COL.  The string is encoded as is, and the encoding ENC is
;;   just used for the prefix, i.e. you are responsible for ensuring
;;   STR is already encoded according to ENC.  The optional argument
;;   nl is the newline separator, defaulting to crlf.

;; This API is compatible with the Gauche library rfc.base64.

(cond-expand
 ((and compiling (not static))
  (declare
   (fixnum)
   (export
    base64-encode base64-encode-string
    base64-decode base64-decode-string
    base64-encode-header)))
 (else))

(define-constant *default-max-col* 76)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; from SRFI-33

(define-inline (%mask size) (- (arithmetic-shift 1 size) 1))
(define-inline (extract-bit-field size pos n)
  (arithmetic-shift (bitwise-and n (arithmetic-shift (%mask size) pos))
                    (- pos)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-constant *outside-char* 99) ; luft-balloons
(define-constant *pad-char* 101)    ; dalmations

(define-constant *base64-decode-table*
  (let ((res (make-vector #x7F *outside-char*)))
    (let lp ((i 0)) ; map letters
      (when (<= i 25)
        (vector-set! res (+ i 65) i)
        (vector-set! res (+ i 97) (+ i 26))
        (lp (+ i 1))))
    (let lp ((i 0)) ; map numbers
      (when (<= i 9) (vector-set! res (+ i 48) (+ i 52)) (lp (+ i 1))))
    ;; extras (be liberal for different common base64 formats)
    (vector-set! res (char->integer #\+) 62)
    (vector-set! res (char->integer #\-) 62)
    (vector-set! res (char->integer #\/) 63)
    (vector-set! res (char->integer #\_) 63)
    (vector-set! res (char->integer #\~) 63)
    (vector-set! res (char->integer #\=) *pad-char*)
    res))

(define-inline (base64-decode-char c)
  (let ((i (char->integer c)))
    (if (>= i #x7F)
      *outside-char*
      (vector-ref *base64-decode-table* i))))

(define (base64-decode-string str)
  (let* ((len (string-length str))
         (res-len (* 3 (arithmetic-shift (+ 3 len) -2)))
         (res (make-string res-len))
         (i 0) (j 0))
    (call-with-current-continuation
     (lambda (return)
       (define (end) (return (if (= j res-len) res (substring res 0 j))))
       (define (next)
         (if (>= i len)
           (end)
           (let ((c (base64-decode-char (string-ref str i))))
             (set! i (+ 1 i))
             (switch c
               (*outside-char* (next))
               (*pad-char* (end))
               (else c)))))
       (define (out i)
         (string-set! res j (integer->char i))
         (set! j (+ j 1)))
       (let lp ()
         (let* ((b1 (next))
                (b2 (next)))
           (out (bitwise-ior (arithmetic-shift b1 2)
                             (extract-bit-field 2 4 b2)))
           (let ((b3 (next)))
             (out (bitwise-ior
                  (arithmetic-shift (extract-bit-field 4 0 b2) 4)
                  (extract-bit-field 4 2 b3)))
             (out (bitwise-ior
                  (arithmetic-shift (extract-bit-field 2 0 b3) 6)
                  (next)))
             (lp))))))))

(define (base64-decode . o)
  (let-optionals* o ((p (current-input-port)))
    (display (base64-decode-string (with-input-from-port p read-string)))))

(define-constant *base64-encode-table*
  (let ((res (make-vector 64)))
    (let lp ((i 0)) ; map letters
      (when (<= i 25)
        (vector-set! res i (integer->char (+ i 65)))
        (vector-set! res (+ i 26) (integer->char (+ i 97)))
        (lp (+ i 1))))
    (let lp ((i 0)) ; map numbers
      (when (<= i 9)
        (vector-set! res (+ i 52) (integer->char (+ i 48)))
        (lp (+ i 1))))
    (vector-set! res 62 #\+)
    (vector-set! res 63 #\/)
    res))

(define (base64-encode-string str)
  (define (enc i) (vector-ref *base64-encode-table* i))
  (let* ((len (string-length str))
         (quot (quotient len 3))
         (rem (- len (* quot 3)))
         (res-len (arithmetic-shift (+ quot (if (zero? rem) 0 1)) 2))
         (limit (- len 2))
         (res (make-string res-len)))
    (let lp ((i 0) (j 0))
      (if (>= i limit)
        (case rem
          ((1)
           (let ((b1 (char->integer (string-ref str i))))
             (string-set! res j (enc (arithmetic-shift b1 -2)))
             (string-set! res (+ j 1) (enc (arithmetic-shift (bitwise-and #b11 b1) 4)))
             (string-set! res (+ j 2) #\=)
             (string-set! res (+ j 3) #\=)))
          ((2)
           (let ((b1 (char->integer (string-ref str i)))
                 (b2 (char->integer (string-ref str (+ i 1)))))
             (string-set! res j (enc (arithmetic-shift b1 -2)))
             (string-set! res (+ j 1) (enc (bitwise-ior
                                            (arithmetic-shift (bitwise-and #b11 b1) 4)
                                            (extract-bit-field 4 4 b2))))
             (string-set! res (+ j 2) (enc (arithmetic-shift (extract-bit-field 4 0 b2) 2)))
             (string-set! res (+ j 3) #\=))))
        (let ((b1 (char->integer (string-ref str i)))
              (b2 (char->integer (string-ref str (+ i 1))))
              (b3 (char->integer (string-ref str (+ i 2)))))
          (string-set! res j (enc (arithmetic-shift b1 -2)))
          (string-set! res (+ j 1) (enc (bitwise-ior
                                         (arithmetic-shift (bitwise-and #b11 b1) 4)
                                         (extract-bit-field 4 4 b2))))
          (string-set! res (+ j 2) (enc (bitwise-ior
                                         (arithmetic-shift (extract-bit-field 4 0 b2) 2)
                                         (extract-bit-field 2 6 b3))))
          (string-set! res (+ j 3) (enc (bitwise-and #b111111 b3)))
          (lp (+ i 3) (+ j 4)))))
    res))

(define (base64-encode . o)
  (let-optionals* o ((p (current-input-port)))
    (display (base64-decode-string (with-input-from-port p read-string)))))

(define (base64-encode-header encoding str . o)
  (define (round4 i) (arithmetic-shift (arithmetic-shift i -2) 2))
  (let-optionals* o ((start-col 0)
                     (max-col *default-max-col*)
                     (nl "\r\n"))
    (let* ((prefix (string-append "=?" encoding "?B?"))
           (prefix-length (+ 2 (string-length prefix)))
           (effective-max-col (round4 (- max-col prefix-length)))
           (first-max-col (round4 (- effective-max-col start-col)))
           (str (base64-encode-string str))
           (len (string-length str)))
      (if (<= len first-max-col)
        (string-append prefix str "?=")
        (string-append
         (if (positive? first-max-col)
           (string-append
            prefix (substring str 0 first-max-col) "?=" nl "\t" prefix)
           "")
         (string-intersperse (string-chop (substring str first-max-col)
                                          effective-max-col)
                             (string-append "?=" nl "\t" prefix))
         "?=")))))
