;; hato-config.scm -- loading and referencing config files
;;
;; Copyright (c) 2005-2008 Alex Shinn.  All rights reserved.
;; BSD-style license: http://synthcode.com/license.txt

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(use srfi-1 srfi-13 extras utils posix hato-db)

(cond-expand
 ((and chicken compiling)
  (declare
   (export
    assq-ref read-from-file alist? conf?
    conf-load conf-load-in-path conf-load-cascaded
    conf-get conf-get-list conf-multi conf-extend
    conf-load-table
    )))
 (else))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Config lists (cascaded lists of alists)

(define (assq-ref ls key . o)
  (let ((cell (assq key ls)))
    (if cell (cdr cell) (and (pair? o) (car o)))))

(define (read-from-file file . opt)
  (condition-case (call-with-input-file file read)
    (exn () (and (pair? opt) (car opt)))))

(define (alist? x)
  (and (list? x) (every pair? x)))

(define (conf? x)
  (and (list? x) (every alist? x)))

(define (conf-load-cascaded cfg-path file)
  (let ((res (conf-load-in-path
              cfg-path
              (if (symbol? file) (symbol->string file) file))))
    (append res (map (lambda (x) (conf-load-cascaded cfg-path x))
                     (conf-get-list (list res) 'inherits)))))

(define (conf-load-in-path cfg-path file)
  (filter
   pair?
   (append-map
    (lambda (dir)
      (conf-load
       (if (eqv? #\/ (string-ref file 0)) file (make-pathname dir file))))
    cfg-path)))

(define (conf-load file)
  (list
   (parameterize ((case-sensitive #f))
     (read-from-file file '()))))

(define (conf-get cfg key . opt)
  (let loop ((ls cfg))
    (if (pair? ls)
        (let ((res (assq key (car ls))))
          (if res
              (let ((val (cdr res)))
                (if (and (pair? val) (null? (cdr val)))
                    (car val)
                    val))
              (loop (cdr ls))))
        (and (pair? opt) (car opt)))))

(define (conf-get-list cfg key . opt)
  (let ((res (conf-get cfg key)))
    (if res
        (if (or (pair? res) (null? res)) res (list res))
        (if (pair? opt) (car opt) '()))))

(define (conf-multi cfg key)
  (append-map (cut assq-ref <> key '()) cfg))

(define (conf-extend alist cfg)
  (if (pair? alist)
      (cons alist cfg)
      cfg))

;; convert a config setting to a table for lookups.  can be either a
;; list, which is converted directly to a hash-table, or a file, which
;; is either opened as a db file (thus reflecting run-times changes) or
;; if a text file is loaded assuming one line per entry
(define (conf-load-table x)
  (cond ((or (pair? x) (null? x))
         (alist->hash-table (map (lambda (x) (cons x #t)) x) string=?))
        ((string? x)
         (and (file-exists? x)
              (if (db-file? x)
                  (open-db x)
                  (conf-load-table
                   (remove
                    string-null?
                    (with-input-from-file x
                      (port-map string-trim-both read-line)))))))
        (else
         #f)))
