;; hato-daemon.scm -- utility to run a daemon
;;
;; Copyright (c) 2008 Alex Shinn.  All rights reserved.
;; BSD-style license: http://synthcode.com/license.txt

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(use extras posix regex tcp tcp-server)

(cond-expand
 ((and chicken compiling)
  (declare (export running-process-id? daemonize daemon-kill)))
 (else))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (process-id-cmdline-file pid)
  (string-append "/proc/" (number->string pid) "/cmdline"))

(define (run-ps pid)
  (call-with-values
      (lambda () (process "ps" (list "-o" "command" (number->string pid))))
    (lambda (in out sub-pid)
      (let* ((header (read-line in))
             (cmdline (read-line in)))
        (and (eof-object? (peek-char in))
             (not (equal? "" cmdline))
             cmdline)))))

(define running-process-id?
  (cond
   ((file-exists? (process-id-cmdline-file (current-process-id)))
    ;; use the proc filesystem
    (lambda (pid . o)
      (condition-case
          (let ((file (process-id-cmdline-file pid)))
            (and (file-exists? file)
                 (or (null? o)
                     (not (car o))
                     (string-search (car o)
                                    (with-input-from-file file read-line)))))
        (exn () #t))))
   ((run-ps (current-process-id))
    ;; try the output of ps
    (lambda (pid . o)
      (let ((cmdline (run-ps pid)))
        (and (string? cmdline)
             (or (null? o) (not (car o)) (string-search (car o) cmdline))))))
   (else
    ;; no proc filesystem and no ps, just assume it's not running
    (lambda (pid . o) #f))))

(define (daemon-kill pid-file #!key (name #f) (notifier (lambda o #f)))
  (let ((pid (and (file-exist? pid-file)
                  (condition-case
                      (call-with-input-file pid-file read)
                    (exn () #f)))))
    (cond
      ((and (number? pid) (running-process-id? pid name))
       (notifier "shutting down server with process ID: ~A" pid)
       (condition-case
        (process-signal pid)
        (exn ()
          (notifier "couldn't SIGTERM process ~A, trying SIGKILL" pid)
          (condition-case
           (process-signal pid signal/kill)
           (exn () (notifier "couldn't SIGKILL process"))))))
      ((number? pid)
       (notifier "removing stale lock file for pid ~A" pid)
       (condition-case (delete-file hato-pid-file)
         (exn ()
           (notifier "couldn't remove stale lock file: ~A"
                     (exception-message exn)))))
      (else
       (notifier "no process to kill")))))

(define (daemonize #!key name pid-file user-id group-id
                   running-handler stale-handler invalid-handler
                   root-thunk init-thunk tcp-port tcp-backlog tcp-host
                   tcp-handler tcp-max-requests tcp-debug?)

  ;; verify there's no process already running
  (if (and pid-file (file-exists? pid-file))
      (let ((pid (string->number (call-with-input-file pid-file read-line))))
        (cond
         ((not (number? pid))
          (if invalid-handler
              (invalid-handler)
              (error "invalid pid file" pid-file)))
         ((running-process-id? pid name)
          (if running-handler
              (running-handler)
              (error "process already running" pid)))
         (else
          (if stale-handler
              (stale-handler)
              (error "stale pid file" pid-file))))))

  ;; fork the process and create a new session
  (if (zero? (process-fork))
      (create-session)
      (exit 0))

  ;; bind the tcp port and do anything needed as root
  (if root-thunk (root-thunk))
  (let ((listener (and tcp-port
                       (tcp-listen tcp-port (or tcp-backlog 4) tcp-host))))

    ;; now switch the user and group
    (cond
     ((zero? (current-user-id))
      (if group-id (set! (current-group-id) group-id))
      (if user-id (set! (current-user-id) user-id))))

    ;; log the pid file
    (if pid-file
        (call-with-output-file pid-file
          (lambda (out) (write (current-process-id) out))))

    ;; run any inits
    (if init-thunk (init-thunk))

    ;; maybe run the tcp-server
    (cond
     (tcp-handler
      (if listener
          ((make-tcp-server listener tcp-handler (or tcp-max-requests 100))
           tcp-debug?)
          (if tcp-port
              (error "couldn't bind to port" tcp-port)
              (error "tcp-handler requires a tcp-port argument"))))
     (listener))

    ))

