;;;; hato-date.scm -- time and date utilities
;;
;; Copyright (c) 2008 Alex Shinn.  All rights reserved.
;; BSD-style license: http://synthcode.com/license.txt

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(use regex regex-case posix srfi-13)

(cond-expand
 ((and chicken compiling)
  (declare
   (export parse-date date<=? date>=? date=? time<=? time>=? time=?)))
 (else))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (wday->number x)
  (case (string->symbol
         (string-downcase
          (if (symbol? x) (symbol->string x) x)))
    ((sun) 0)
    ((mon) 1)
    ((tue) 2)
    ((wed) 3)
    ((thu) 4)
    ((fri) 5)
    ((sat) 6)
    (else #f)))

(define (mon->number x)
  (case (string->symbol
         (string-downcase
          (if (symbol? x) (symbol->string x) x)))
    ((jan) 1)
    ((feb) 2)
    ((mar) 3)
    ((apr) 4)
    ((may) 5)
    ((jun) 6)
    ((jul) 7)
    ((aug) 8)
    ((sep) 9)
    ((oct) 10)
    ((nov) 11)
    ((dec) 12)
    (else #f)))

(define (tz->number x)
  (let ((m (and (string? x) (string-match "^([-+]?\\d{2})(\\d+)" x))))
    (if m
        (* 60 (+ (* 60 (string->number (cadr m))) (string->number (caddr m))))
        0)))

(define (make-date seconds minutes hours mday month year wday yday dstflag tz)
  (seconds->local-time
   (local-time->seconds
    (vector seconds minutes hours mday month year wday yday dstflag tz))))

;; universal, loose parser
(define parse-date
  (let ()
    (define (i x) (or (and (string? x) (string->number x)) 0))
    (define (save . args) (apply string-append "(" (append args '(")"))))
    (define (group . args) (apply string-append "(?:" (append args '(")"))))
    (define (rx-or ls) (save (string-join ls "|")))
    (define (? . args) (string-append (apply group args) "?"))
    (define (abbrev-rx x n)
      (let ((str (->string x)))
        (string-append (string-take str n) (? (string-drop str n)))))
    (define (rx-abbrev-ls ls n)
      (rx-or (map (cut abbrev-rx <> n) ls)))
    (define wday-en '(Sunday Monday Tuesday Wednesday Thursday Friday Saturday))
    (define rx-wday-en (rx-abbrev-ls wday-en 3))
    (define mon-en '(January February March April May June July
                             August September October November December))
    (define rx-mon-en (rx-abbrev-ls mon-en 3))
    (define sep "\\s*[-_,;:/]?\\s*")
    (define d2 "(\\d{1,2})")
    (define d4 "(\\d{1,4})")
    (define rx-tz
      (? (? (rx-or '("[-+]\\d{2}[013][05]" "\\w+/\\w+"))
            (? "\\s*\\(?[A-Za-z]{3}\\)?\\s*"))))
    (define rx-mag
      (regexp
       (string-append
        "^" d4 sep d2 sep d2
        (? (? "T") sep d2 sep d2 sep d2 sep (? sep rx-tz)))))
    (define rx-822
      (regexp
       (string-append (? rx-wday-en sep) d2 sep rx-mon-en sep d4
                      (? sep d2 sep d2 sep d2) (? sep rx-tz))
       #t))
    (define rx-unix
      (regexp
       (string-append (? rx-wday-en sep) rx-mon-en sep d2
                      (? sep d2 sep d2 sep d2) (? sep rx-tz)
                      sep d4)))

    (lambda (str)
      (regex-case
       (string-trim str)
       ;; Put the formats with named weekdays/months first since if
       ;; they occur we can be confident of a match.
       ;;
       ;; RFC-822 style (~a, ~d ~b ~Y ~H:~M:~S ~z)
       (rx-822 (_ w d m y h mi ss tz)
               (make-date (i ss) (i mi) (i h) (i d) (- (mon->number m) 1)
                          (- (i y) 1900) 0 0 #f (tz->number tz)))
       ;; Unix style
       (rx-unix (_ w m d h mi ss tz y)
                (make-date (i ss) (i mi) (i h) (i d) (- (mon->number m) 1)
                           (- (i y) 1900) 0 0 #f (tz->number tz)))
       ;; YYYY-MM-DD style
       (rx-mag (_ y m d h mi ss tz)
               (make-date (i ss) (i mi) (i h) (i d) (- (i m) 1) (- (i y) 1900)
                          0 0 #f (tz->number tz)))
       (else
        ;;(warning "unrecognized date format: ~S" str)
        #f)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-inline (date-value a)
  (+ (* 10000 (vector-ref a 5))
     (* 100 (vector-ref a 4))
     (vector-ref a 3)))

(define (date<=? a b)
  (let ((a (if (number? a) (seconds->local-time a) a))
        (b (if (number? b) (seconds->local-time b) b)))
    (<= (date-value a) (date-value b))))

(define (date>=? a b)
  (let ((a (if (number? a) (seconds->local-time a) a))
        (b (if (number? b) (seconds->local-time b) b)))
    (<= (date-value a) (date-value b))))

(define (date=? a b)
  (let ((a (if (number? a) (seconds->local-time a) a))
        (b (if (number? b) (seconds->local-time b) b)))
    (= (date-value a) (date-value b))))

(define (time<=? a b)
  (<= (if (number? a) a (local-time->seconds a))
      (if (number? b) b (local-time->seconds b))))

(define (time>=? a b)
  (>= (if (number? a) a (local-time->seconds a))
      (if (number? b) b (local-time->seconds b))))

(define (time=? a b)
  (= (if (number? a) a (local-time->seconds a))
     (if (number? b) b (local-time->seconds b))))

