;;;; html-parser.scm -- SSAX-like tree-folding html parser
;;
;; Copyright (c) 2003-2008 Alex Shinn.  All rights reserved.
;; BSD-style license: http://synthcode.com/license.txt

;; This is intended as a permissive HTML parser for people who prefer
;; the interface described in Oleg Kiselyov's SSAX parser.  It correctly
;; handles all invalid HTML, inserting "virtual" starting and closing
;; tags as needed to maintain the proper tree structure needed for the
;; foldts down/up logic.  A major goal of this parser is bug-for-bug
;; compatibility with the way common web browsers parse HTML.

;; Procedure: make-html-parser . keys
;;   Returns a procedure of two arguments, and initial seed and an
;;   optional input port, which parses the HTML document from the port
;;   with the callbacks specified in the DSSSL-specified KEYS.  The
;;   following callbacks are recognized:
;;
;;   START TAG ATTRS SEED VIRTUAL?
;;       fdown in foldts, called when a start-tag is encountered.
;;     TAG:         tag name
;;     ATTRS:       tag attributes as a alist
;;     SEED:        current seed value
;;     VIRTUAL?:    #t iff this start tag was inserted to fix the HTML tree
;;
;;   END TAG ATTRS PARENT-SEED SEED VIRTUAL?
;;       fup in foldts, called when an end-tag is encountered.
;;     TAG:         tag name
;;     ATTRS:       tag attributes of the corresponding start tag
;;     PARENT-SEED: parent seed value (i.e. seed passed to the start tag)
;;     SEED:        current seed value
;;     VIRTUAL?:    #t iff this end tag was inserted to fix the HTML tree
;;
;;   TEXT TEXT SEED
;;       fhere in foldts, called when any text is encountered.  May be
;;       called multiple times between a start and end tag, so you need
;;       to string-append yourself if desired.
;;     TEXT:        entity-decoded text
;;     SEED:        current seed value
;;
;;   COMMENT TEXT SEED
;;       fhere on comment data
;;
;;   DECL NAME ATTRS SEED
;;       fhere on declaration data
;;       
;;   PROCESS LIST SEED
;;       fhere on process-instruction data
;;
;;   In addition, entity-mappings may be overriden with the ENTITIES:
;;   keyword.

;; Procedure: html->sxml [port]
;;   Returns the SXML representation of the document from PORT, using
;;   the default parsing options.

;; Procedure: html-strip [port]
;;   Returns a string representation of the document from PORT with all
;;   tags removed.  No whitespace reduction or other rendering is done.

;; Example:
;;
;;   The parser for html-strip could be defined as:
;;
;; (make-html-parser
;;   start: (lambda (tag attrs seed virtual?) seed)
;;   end:   (lambda (tag attrs parent-seed seed virtual?) seed)
;;   text:  (lambda (text seed) (display text)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(cond-expand
 ((and compiling (not static))
  (declare
   (fixnum)
   (export make-html-parser html->sxml html-strip
           sxml-display-as-html sxml->html
           html-escape html-escape-attr
           html-attr->string html-tag->string)))
 (else))

(define-macro (let-keywords* ls params . body)
  `(apply (lambda (#!key ,@params) ,@body) ,ls))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; text parsing utils

(define (read-while pred . o)
  (let ((p (if (pair? o) (car o) (current-input-port))))
    (with-output-to-string
      (lambda ()
        (let loop ((c (peek-char p)))
          (if (and (not (eof-object? c)) (pred c))
            (begin
              (display (read-char p))
              (loop (peek-char p)))))))))

(define (read-until pred . o)
  (let ((p (if (pair? o) (car o) (current-input-port))))
    (with-output-to-string
      (lambda ()
        (let loop ((c (peek-char p)))
          (if (not (or (eof-object? c) (pred c)))
            (begin
              (display (read-char p))
              (loop (peek-char p)))))))))

(define (read-until-string str . o)
  (define (make-queue) (cons '() '())) ; (first . last)
  (define (enqueue! q x)
    (let ((cell (list x)))
      (if (null? (car q))
        (set-car! q cell)
        (set-cdr! (cdr q) cell))
      (set-cdr! q cell)))
  (define (dequeue! q)
    (let* ((cell (car q)) (rest (cdr cell)))
      (set-car! q rest)
      (if (null? rest) (set-cdr! q '()))
      (car cell)))
  (let ((p (if (pair? o) (car o) (current-input-port))))
    (with-output-to-string
      (lambda ()
        (let ((target (make-queue))
              (current (make-queue))
              (len (string-length str)))
          (let loop ((i 0))
            (if (< i len)
              (begin
                (enqueue! target (string-ref str i))
                (loop (+ i 1)))))
          (let loop ((c (peek-char p)) (i 0))
            (if (not (or (eof-object? c) (>= i len)))
              (begin
                (enqueue! current (read-char p))
                (loop (peek-char p) (+ i 1)))))
          (let loop ((c (peek-char p)))
            (if (not (equal? (car current) (car target)))
              (begin
                (enqueue! current (read-char p))
                (display (dequeue! current))
                (loop (peek-char p))))))))))

(define skip-while read-while)
(define skip-until read-until)

(define skip-whitespace (lambda x (apply read-while char-whitespace? x)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; html-specific readers

(define (char-alphanumeric? c)
  (or (char-alphabetic? c) (char-numeric? c)))

(define read-identifier (lambda x (apply read-while char-alphanumeric? x)))

(define read-integer (lambda x (apply read-while char-numeric? x)))

(define (read-quoted . o)
  (apply read-until (lambda (c) (eqv? c #\")) o))

(define (read-name-or-quoted . o)
  (let ((p (if (pair? o) (car o) (current-input-port))))
    (if (eqv? (peek-char p) #\")
      (begin
        (read-char p)
        (let ((res (read-quoted p)))
          (read-char p)
          res))
      (read-until char-whitespace? p))))

(define (read-pi . o)
  (let ((p (if (pair? o) (car o) (current-input-port))))
    (let loop ((c (peek-char p)) (res (list)))
      (cond
        ((or (eof-object? c) (eqv? c #\>))
         (read-char p)
         (reverse res))
        (else
         (cond
           ((char-whitespace? c)
            (skip-whitespace p)
            (loop (peek-char p) res))
           ((eqv? c #\")
            (read-char p)
            (let ((res (cons (read-quoted p) res)))
              (loop (peek-char p) res)))
           ((char-alphanumeric? c)
            (let ((res (cons (read-identifier p) res)))
              (loop (peek-char p) res)))
           (else
            (read-char p)
            (loop (peek-char p) res))))))))

(define read-comment (lambda x (apply read-until-string "-->" x)))

(define (read-start . o)
  (let ((p (if (pair? o) (car o) (current-input-port))))
    (skip-whitespace p)
    (let ((tag (string->symbol (read-until (disjoin char-whitespace?
                                                    (lambda (c) (eqv? c #\>)))
                                           p))))
      (let loop ((c (peek-char p)) (attrs '()))
        (if (or (eof-object? c) (eqv? c #\>))
          (begin (read-char p) (list tag (reverse attrs)))
          (let ((name (read-until (disjoin char-whitespace?
                                           (lambda (c) (eqv? c #\>))
                                           (lambda (c) (eqv? c #\=)))
                                  p)))
            (skip-whitespace p)
            (if (string=? name "")
              (loop (peek-char p) attrs)
              (let ((name (string->symbol name)))
                (cond
                  ((eqv? (peek-char p) #\=)
                   (read-char p)
                   (skip-whitespace p)
                   (let ((a (cons (cons name (read-name-or-quoted p)) attrs)))
                     (loop (peek-char p) a)))
                  (else
                   (let ((attrs (cons (cons name #t) attrs)))
                     (loop (peek-char p) attrs))))))))))))

(define (read-end . o)
  (car (apply read-start o)))

(define read-decl read-start)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; the parser

(define *default-entities*
  '(("amp" . "&") ("quot" . "\"") ("lt" . "<")
    ("gt" . ">")  ("apos" . "'")  ("nbsp" . " ")))

(define (get-entity entities name)
  (cond
    ;;((string->number name) => (compose string integer->char))
    ((assoc name entities) => cdr)
    (else #f)))

(define *tag-levels*
  '(html (head body) table tr (th td) p (b i u)))

(define *unnestables*
  '(p li td tr))

(define (tag-level tag-levels tag)
  (let lp ((ls tag-levels) (i 0))
    (if (null? ls)
      (+ i 1000)
      (if (if (pair? (car ls))
            (memq tag (car ls))
            (eq? tag (car ls)))
        i
        (lp (cdr ls) (+ i 1))))))

(define (read-html-token . o)
  (let ((p (if (pair? o) (car o) (current-input-port))))
    (let ((c (peek-char p)))
      (if (eof-object? c)
        (cons 'eof c)
        (case c
          ((#\<)
           (read-char p)
           (case (peek-char p)
             ((#\!)
              (read-char p)
              (cond
                ((eqv? #\[ (peek-char p))
                 (read-char p)
                 (let lp ((check '(#\C #\D #\A #\T #\A #\[))
                          (acc '(#\[ #\! #\<)))
                   (cond
                     ((null? check)
                      (cons 'text (read-until-string "]]>" p)))
                     ((let ((c (peek-char p)))
                        (and (not (eof-object? c)) (char-ci=? c (car check))))
                      (lp (cdr check) (cons (read-char p) acc)))
                     (else
                      (cons 'text (list->string (reverse acc)))))))
                ((and (eqv? #\- (peek-char p))
                      (read-char p)
                      (eqv? #\- (peek-char p)))
                 (cons 'comment (read-comment p)))
                (else
                 (cons 'decl (read-decl p)))))
             ((#\?)
              (read-char p)
              (cons 'process (read-pi p)))
             ((#\/)
              (read-char p)
              (cons 'end (read-end p)))
             (else
              ;; start tags must immediately be followed by an
              ;; alphabetic charater, or we just treat the < as text
              (if (char-alphabetic? (peek-char p))
                (cons 'start (read-start p))
                (cons 'text "<")))))
          ((#\&)
           (read-char p)
           (if (eqv? (peek-char p) #\#)
             (begin
               (read-char p)
               (let* ((str (read-integer p))
                      (num (string->number str)))
                 (if (and num (eqv? (peek-char p) #\;))
                   (begin (read-char p) (cons 'entity num))
                   (cons 'text (string-append "&#" str)))))
             (let ((name (read-identifier p)))
               (if (eqv? (peek-char p) #\;)
                 (begin (read-char p) (cons 'entity name))
                 (cons 'text (string-append "&" name))))))
          (else
           (cons 'text (read-until (lambda (c) (or (eqv? c #\<) (eqv? c #\&)))
                                   p))))))))

(define (make-html-parser . o)
  (let-keywords* o
      ((start (lambda (t a s v) s)) (end (lambda (t a p s v) s))
       (text (lambda (t s) s)) (decl (lambda (t a s) s))
       (process (lambda (t s) s)) (comment (lambda (t s) s))
       (entities *default-entities*)
       (tag-levels *tag-levels*)
       (unnestables *unnestables*)
       (entity
        (lambda (t s)
          (text (if (number? t)
                  (string (integer->char t))
                  (or (get-entity entities t)
                      (string-append "&" t ";")))
                s))))
    (lambda (seed . o)
      (let ((p (if (pair? o) (car o) (current-input-port))))
        (let lp ((tok (read-html-token p)) (seed seed) (seeds '()) (tags '()))
          (case (car tok)
            ((eof) ; close all open tags
             (let lp ((t tags) (s seeds) (seed seed))
               (if (null? t)
                 seed
                 (lp (cdr t) (cdr s)
                     (end (caar t) (cadar t) (car s) seed 'eof)))))
            ((start)
             (let ((tag (cadr tok)))
               (if (and (pair? tags) (eq? tag (caar tags))
                        (memq tag unnestables))
                 ;; <p> ... <p> implies siblings, not nesting
                 (let ((seed2
                        (end tag (cadar tags) (car seeds) seed 'sibling)))
                   (lp (read-html-token p)
                       (start tag (caddr tok) seed #f)
                       (cons seed2 (cdr seeds))
                       (cons (cdr tok) (cdr tags))))
                 (lp (read-html-token p)
                     (start tag (caddr tok) seed #f)
                     (cons seed seeds)
                     (cons (cdr tok) tags)))))
            ((end)
             (if (and (pair? tags) (eq? (cdr tok) (caar tags)))
               (lp (read-html-token p)
                   (end (cdr tok) (cadar tags) (car seeds) seed #f)
                   (cdr seeds)
                   (cdr tags))
               (let ((this-level (tag-level tag-levels (cdr tok)))
                     (expected-level
                      (if (pair? tags) (tag-level tag-levels (caar tags)) -1)))
                 (cond
                   ((< this-level expected-level)
                    ;; higher-level tag, forcefully close preceding tags
                    (lp tok
                        (end (caar tags) (cadar tags) (car seeds) seed
                             'parent-closed)
                        (cdr seeds)
                        (cdr tags)))
                   ((and (= this-level expected-level) (pair? (cdr tags)))
                    ;; equal, interleave (close prec tag, close this,
                    ;; re-open prec)
                    ;; <b><i></b> => <b><i></i></b><i>
                    ;;                     ^^^^    ^^^
                    ;; XXXX handle backups > 1 here
                    (let* ((seed2 (end (caar tags) (cadar tags)
                                       (car seeds) seed 'interleave))
                           (seed3 (end (caadr tags) (cadadr tags)
                                       (cadr seeds) seed2 #f)))
                      (lp (read-html-token p)
                          (start (caar tags) (cadar tags) seed3 'interleave)
                          (cons seed3 (cddr seeds))
                          (cons (car tags) (cddr tags)))))
                   (else
                    ;; spurious end for a lower-level tag, add imaginary start
                    (let* ((seed2 (start (cdr tok) '() seed 'no-start))
                           (seed3 (end (cdr tok) '() seed seed2 #f)))
                      (lp (read-html-token p) seed3 seeds tags)))))))
            ((text)
             (lp (read-html-token p) (text (cdr tok) seed) seeds tags))
            ((entity)
             (lp (read-html-token p) (entity (cdr tok) seed) seeds tags))
            ((comment)
             (lp (read-html-token p) (comment (cdr tok) seed) seeds tags))
            ((decl)
             (lp (read-html-token p)
                 (decl (car tok) (cdr tok) seed) seeds tags))
            ((process)
             (lp (read-html-token p) (process (cdr tok) seed) seeds tags))
            (else
             (error "invalid token: " tok))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; simple conversions

(define html->sxml
  (let ((parse
         (make-html-parser
          start: (lambda (tag attrs seed virtual?) '())
          end:   (lambda (tag attrs parent-seed seed virtual?)
                   `((,tag ,@(if (pair? attrs)
                               `((@ ,@attrs) ,@(reverse seed))
                               (reverse seed)))
                     ,@parent-seed))
          text:  (lambda (text seed) (cons text seed))
          )))
    (lambda o
      (reverse (apply parse '() o)))))

(define (html-escape-attr str)
  str)

(define (html-attr->string attr)
  (string-append (symbol->string (car attr))
                 "=\"" (html-escape-attr (cdr attr)) "\""))

(define (html-tag->string tag attrs)
  (let lp ((ls attrs) (res (list (symbol->string tag) "<")))
    (if (null? ls)
      (apply string-append (reverse (cons ">" res)))
      (lp (cdr ls) (cons (html-attr->string (car ls)) (cons " " res))))))

(define (html-display-escaped-string str)
  (let ((start 0)
        (end (string-length str)))
    (let lp ((from start) (to start))
      (if (>= to end)
          (display (substring str from to))
          (let ((c (string-ref str to)))
            (cond
             ((eq? c #\<)
              (display (substring str from to))
              (display "&lt;")
              (let ((next (+ to 1)))
                (lp next next)))
             ((eq? c #\&)
              (display (substring str from to))
              (display "&amp;")
              (let ((next (+ to 1)))
                (lp next next)))
             (else
              (lp from (+ to 1)))))))))

(define (html-escape str)
  (with-output-to-string
    (lambda () (html-display-escaped-string str))))

(define (sxml-display-as-html sxml . o)
  (cond
    ((pair? sxml)
     (let ((tag (car sxml)))
       (if (symbol? tag)
         (let ((rest (cdr sxml)))
           (cond
             ((and (pair? rest)
                   (pair? (car rest))
                   (eq? '@ (caar rest)))
              (display (html-tag->string tag (cdar rest)))
              (for-each sxml-display-as-html (cdr rest))
              (display "</") (display tag) (display ">"))
             (else
              (display (html-tag->string tag '()))
              (for-each sxml-display-as-html rest)
              (display "</") (display tag) (display ">"))))
         (for-each sxml-display-as-html sxml))))
    ((null? sxml))
    (else (html-display-escaped-string sxml))))

(define (sxml->html sxml . o)
  (with-output-to-string
    (lambda () (apply sxml-display-as-html sxml o))))

;; just strips tags, no whitespace handling or formatting
(define html-strip
  (let ((parse
         (make-html-parser
          start: (lambda (tag attrs seed virtual?) seed)
          end:   (lambda (tag attrs parent-seed seed virtual?) seed)
          text:  (lambda (text seed) (display text))
          )))
    (lambda o
      (with-output-to-string
        (lambda () (apply parse (cons #f #f) o))))))
